# -*- coding: utf-8 -*-
# file: config.py
# time: 23:04 26/04/2024
# author: YANG, HENG <hy345@exeter.ac.uk> (杨恒)
# github: https://github.com/yangheng95
# huggingface: https://huggingface.co/yangheng
# google scholar: https://scholar.google.com/citations?user=NPq5a_0AAAAJ&hl=en
# Copyright (C) 2019-2024. All Rights Reserved.

import os

import torch

from omnigenome import (
    ClassificationMetric,
    AutoBenchConfig,
    OmniGenomeModel,
    OmniGenomeDataset,
    OmniGenomePooling
)
from transformers import BatchEncoding


class Model(OmniGenomeModel):
    def __init__(self, config_or_model, tokenizer, *args, **kwargs):
        super().__init__(config_or_model, tokenizer, *args, **kwargs)
        self.metadata["model_name"] = self.__class__.__name__
        # self.softmax = torch.nn.Softmax(dim=-1)
        self.classifier = torch.nn.Linear(
            self.config.hidden_size, self.config.num_labels
        )
        self.loss_fn = torch.nn.BCEWithLogitsLoss()
        self.pooler = OmniGenomePooling(self.config)
        self.model_info()

    def forward(self, inputs):
        last_hidden_state = self.last_hidden_state_forward(inputs)
        last_hidden_state = self.dropout(last_hidden_state)
        last_hidden_state = self.activation(last_hidden_state)

        if self.pooler._is_causal_lm():
            logits = self.classifier(last_hidden_state)
            pad_token_id = getattr(self.config, "pad_token_id", -100)
            sequence_lengths = inputs["input_ids"].ne(pad_token_id).sum(dim=1) - 1
            logits = logits[
                torch.arange(inputs["input_ids"].size(0), device=logits.device),
                sequence_lengths,
            ]
        else:
            last_hidden_state = self.pooler(inputs, last_hidden_state)

            logits = self.classifier(last_hidden_state)

        outputs = {"logits": logits, "last_hidden_state": last_hidden_state}
        return outputs

    def predict(self, sequence_or_inputs, **kwargs):
        if not isinstance(sequence_or_inputs, BatchEncoding) and not isinstance(
                sequence_or_inputs, dict
        ):
            inputs = self.tokenizer(
                sequence_or_inputs,
                padding=kwargs.pop("padding", "max_length"),
                max_length=kwargs.pop("max_length", 512),
                truncation=True,
                return_tensors="pt",
                **kwargs,
            )
        else:
            inputs = sequence_or_inputs
        inputs = inputs.to(self.model.device)

        with torch.no_grad():
            outputs = self(inputs)
        logits = outputs["logits"]
        last_hidden_state = outputs["last_hidden_state"]

        predictions = []
        for i in range(logits.shape[0]):
            prediction = logits[i].ge(0.5).to(torch.int).cpu().numpy()

            predictions.append(prediction)

        outputs = {
            "predictions": predictions,
            "logits": logits,
            "last_hidden_state": last_hidden_state,
        }

        return outputs

    # def inference(self, sequence_or_inputs, **kwargs):
    #     if not isinstance(sequence_or_inputs, BatchEncoding) and not isinstance(
    #         sequence_or_inputs, dict
    #     ):
    #         inputs = self.tokenizer(
    #             sequence_or_inputs,
    #             padding=kwargs.pop("padding", "max_length"),
    #             max_length=kwargs.pop("max_length", 512),
    #             truncation=True,
    #             return_tensors="pt",
    #             **kwargs
    #         )
    #     else:
    #         inputs = sequence_or_inputs
    #     inputs = inputs.to(self.model.device)
    #
    #     with torch.no_grad():
    #         outputs = self(inputs)
    #     logits = outputs["logits"]
    #     last_hidden_state = outputs["last_hidden_state"]
    #
    #     predictions = []
    #     for i in range(logits.shape[0]):
    #         predictions.append(
    #             self.config.id2label.get(logits[i].argmax(dim=-1).item(), "")
    #         )
    #
    #     if not isinstance(sequence_or_inputs, list):
    #         outputs = {
    #             "predictions": predictions[0],
    #             "logits": logits[0],
    #             "last_hidden_state": last_hidden_state[0],
    #         }
    #     else:
    #         outputs = {
    #             "predictions": predictions,
    #             "logits": logits,
    #             "last_hidden_state": last_hidden_state,
    #         }
    #
    #     return outputs

    def loss_function(self, logits, labels):
        labels = labels[labels != -100]
        loss = self.loss_fn(logits.view(-1), labels.view(-1).to(torch.float32))
        return loss


class Dataset(OmniGenomeDataset):
    def __init__(self, data_source, tokenizer, max_length=None, **kwargs):
        super(Dataset, self).__init__(data_source, tokenizer, max_length, **kwargs)

        for key, value in kwargs.items():
            self.metadata[key] = value

    def prepare_input(self, instance, **kwargs):
        labels = None
        if isinstance(instance, str):
            sequence = instance
        elif isinstance(instance, dict):
            sequence = (
                instance.get("seq", None)
                if "seq" in instance
                else instance.get("sequence", None)
            )
            label = instance.get("label", None)
            labels = instance.get("labels", None)
            labels = labels if labels is not None else label
        else:
            raise Exception("Unknown instance format.")

        tokenized_inputs = self.tokenizer(
            sequence,
            padding="do_not_pad",
            truncation=True,
            max_length=self.max_length,
            return_tensors="pt",
            **kwargs,
        )
        for col in tokenized_inputs:
            tokenized_inputs[col] = tokenized_inputs[col].squeeze()

        if labels is not None:
            tokenized_inputs["labels"] = torch.tensor(labels)
            tokenized_inputs["labels"] = torch.tensor(tokenized_inputs["labels"])
        return tokenized_inputs


# Hyperparameters
config_dict = {
    "task_name": "chromatin_access.oryza_sativa_ZS97_RS2",
    "task_type": "multi_label_sequence_classification",
    "label2id": None,  # For Sequence Classification
    "num_labels": 15,  # For Sequence Classification
    "epochs": 20,
    "learning_rate": 2e-5,
    "weight_decay": 0,
    "batch_size": 32,
    "max_length": 512,  # "max_length": 1024 for some models
    "seeds": [45],
    "compute_metrics": ClassificationMetric(ignore_y=-100, average="macro").f1_score,
    "train_file": f"{os.path.dirname(__file__)}/train.json",
    "test_file": f"{os.path.dirname(__file__)}/test.json",
    "valid_file": f"{os.path.dirname(__file__)}/valid.json"
    if os.path.exists(f"{os.path.dirname(__file__)}/valid.json")
    else None,
    "dataset_cls": Dataset,  # For your custom dataset preparation
    "model_cls": Model,
}

bench_config = AutoBenchConfig(config_dict)
